#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#define RR_QUANTUM 2
#define CNTXT_SWITCH 1

enum pstate {
    WAITING,
    READY,
    FINISHED
};

struct pinfo {
    int id;
    int arrival_time;
    int execution_time;
    int priority;
    
    int wait_time;
    int completion_time;
    enum pstate state;

    struct pinfo * next_pinfo;
};

struct perf_info {
    int total_time;
    int total_nr_ctxt_switch;
    int total_time_ctxt_switch;
};

void print_perf(struct perf_info * perf) {
    printf("Total time: %d\n", perf->total_time);
    printf("Total number of context switches: %d\n", perf->total_nr_ctxt_switch);
    printf("Total time spent on context switching: %d\n", perf->total_time_ctxt_switch);
}

void print_pinfo(struct pinfo * info) {
    printf("<Process {\n");
    printf("  PID: %d\n", info->id);
    printf("  Arrival time: %d\n", info->arrival_time);
    printf("  Execution time: %d\n", info->execution_time);
    printf("  Priority: %d\n", info->priority);
    printf("  Wait time: %d\n", info->wait_time);
    printf("  Completion time: %d\n", info->completion_time);
    printf("  NEXT -> %p\n", info->next_pinfo);
    printf("}>\n");
}

void print_processes(struct pinfo * processes) {
    while (processes != NULL) {
        print_pinfo(processes);
        processes = processes->next_pinfo;
    }
}

struct pinfo * create_process(int id, int arrival_time, int execution_time, int priority) {
    struct pinfo * info = malloc(sizeof(struct pinfo));
    info->id = id;
    info->arrival_time = arrival_time;
    info->execution_time = execution_time;
    info->priority = priority;
    info->wait_time = 0;
    info->completion_time = 0;
    info->state = WAITING;
    info->next_pinfo = NULL;
    return info;
}

struct perf_info schedule_FCFS(struct pinfo * processes) {
    int current_time = 0;
    struct pinfo * process = processes;

    struct perf_info perf = {0, 0, 0};
    
    while (process != NULL) {
        int wait_time = current_time - process->arrival_time;
        if (wait_time < 0) {
            wait_time = 0;
        }
        process->wait_time = wait_time;
        process->completion_time = process->execution_time + process->wait_time;
        current_time = process->arrival_time + process->completion_time;

        process = process->next_pinfo;
    }
    perf.total_time = current_time;
    
    return perf;
}


struct pinfo * read_file() {
    FILE * file = fopen("tasks.csv", "r");
    unsigned long buf_size = sizeof(char) * 64;
    char * line = (char *) malloc(buf_size);
    char * pid_str;
    char * arrival_str;
    char * execution_str;
    char * prio_str;

    struct pinfo * first = NULL;
    struct pinfo * last = NULL;
    struct pinfo * process;

    while (fgets(line, buf_size, file)) {
        pid_str = strtok(line, " ");
        arrival_str = strtok(NULL, " ");
        execution_str = strtok(NULL, " ");
        prio_str = strtok(NULL, " ");

        process = create_process(
            atoi(pid_str),
            atoi(arrival_str),
            atoi(execution_str),
            atoi(prio_str)
        );

        // If linked list not initialized
        if (first == NULL) {
            first = process;
        }

        // If there is an element in list
        if (last != NULL) {
            last->next_pinfo = process;
        }
        last = process;

        print_pinfo(process);
    }
    free(line);
    fclose(file);
    return first;
}

void write_file(struct pinfo * process, struct perf_info * perf) {
    // TODO
}

void free_processes(struct pinfo * next) {
    struct pinfo * cur;
    while (next != NULL) {
        cur = next;
        next = cur->next_pinfo;
        free(cur);
    }
}

int main() {
    struct pinfo * processes = read_file();

    struct perf_info perf = schedule_FCFS(processes);

    //write_file(processes, &perf);
    print_processes(processes);
    print_perf(&perf);

    free_processes(processes);

    return 0;
}